#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Geoid Module

In Development
"""

import numpy as np
from scipy import interpolate
from geodepy.conversions import dms2dd


# Define test grid points
nvals = np.array([(dms2dd(-31.51), dms2dd(133.48), -8.806),
                  (dms2dd(-31.51), dms2dd(133.49), -8.743),
                  (dms2dd(-31.52), dms2dd(133.48), -8.870),
                  (dms2dd(-31.52), dms2dd(133.49), -8.805)])

lat = dms2dd(-31.515996736)
long = dms2dd(133.483540489)





#!/usr/bin/env python3

from math import radians, sin, cos, sqrt, atan2, degrees
from numpy import matrix


def rotation_matrix(lat, lon):
    """Returns the 3x3 rotation matrix for a given latitude and longitude
    (given in decimal degrees)
    See Section 4.2.3 of the DynaNet User's Guide v3.3
    """
    (rlat, rlon) = (radians(lat), radians(lon))
    rot_matrix = matrix(
        [[-sin(rlon), -sin(rlat)*cos(rlon), cos(rlat)*cos(rlon)],
         [cos(rlon), -sin(rlat)*sin(rlon), cos(rlat)*sin(rlon)],
         [0.0, cos(rlat), sin(rlat)]]
    )
    return rot_matrix


def vcv_cart2local(vcv_cart, lat, lon):
    """Transforms a 3x3 VCV from the Cartesian to the local reference frame
        See Section 4.4.1 of the DynaNet User's Guide v3.3
    """
    rot_matrix = rotation_matrix(lat, lon)
    vcv_local = rot_matrix.transpose() * vcv_cart * rot_matrix
    return vcv_local


def error_ellipse(vcv):
    """Calculate the semi-major axis, semi-minor axis, and the orientation of
    the error ellipse calculated from a 3x3 VCV
    See Section 7.3.3.1 of the DynaNet User's Guide v3.3
    """
    z = sqrt((vcv[0, 0] - vcv[1, 1])**2 + 4 * vcv[0, 1]**2)
    a = sqrt(0.5 * (vcv[0, 0] + vcv[1, 1] + z))
    b = sqrt(0.5 * (vcv[0, 0] + vcv[1, 1] - z))
    orientation = 90 - degrees(0.5 * atan2((2 * vcv[0, 1]),
                                     (vcv[0, 0] - vcv[1, 1])))

    return a, b, orientation


def circ_hz_pu(a, b):
    """Calculate the circularised horizontal PU form the semi-major and
    semi-minor axes
    """
    q0 = 1.960790
    q1 = 0.004071
    q2 = 0.114276
    q3 = 0.371625
    c = b / a
    k = q0 + q1 * c + q2 * c**2 + q3 * c**3
    r = a * k

    return r



#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Geodesy Module
"""

import os
import csv
from math import (pi, degrees, radians, sqrt, sin,
                  cos, tan, asin, atan, atan2)
import numpy as np
from geodepy.convert import dd2dms, dms2dd
from geodepy.constants import grs80


def enu2xyz(lat, long, east, north, up):
    """
    function to convert a vector in a local east, north, up reference frame to a vector in a
    cartesian x, y, z reference frame
    :param lat: latitude in decimal degrees
    :param long: longitude in decimal degrees
    :param east: in metres
    :param north: in metres
    :param up: in metres
    :return: x, y, z in metres
    """
    lat = radians(lat)
    long = radians(long)
    # Create ENU Vector
    enu = np.array([[east],
                    [north],
                    [up]])
    # Create Rotation Matrix
    rotate = np.array([[-sin(long), -sin(lat)*cos(long), cos(lat)*cos(long)],
                       [cos(long), -sin(lat)*sin(long), cos(lat)*sin(long)],
                       [0, cos(lat), sin(lat)]])
    xyz = np.dot(rotate, enu)
    # Assign to separate variables
    x = float(xyz[0])
    y = float(xyz[1])
    z = float(xyz[2])
    return x, y, z


def xyz2enu(lat, long, x, y, z):
    """
    function to convert a vector in a cartesian x, y, z reference frame to a vector in a
    local east, north, up reference frame
    :param lat: latitude in decimal degrees
    :param long: longitude in decimal degrees
    :param x: in metres
    :param y: in metres
    :param z: in metres
    :return: east, north, up in metres
    """
    lat = radians(lat)
    long = radians(long)
    # Create XYZ Vector
    xyz = np.array([[x],
                    [y],
                    [z]])
    # Create Rotation Matrix
    rotate = np.array([[-sin(long), cos(long), 0],
                       [-sin(lat)*cos(long), -sin(lat)*sin(long), cos(lat)],
                       [cos(lat)*cos(long), cos(lat)*sin(long), sin(lat)]])
    enu = np.dot(rotate, xyz)
    # Assign to separate variables
    east = float(enu[0])
    north = float(enu[1])
    up = float(enu[2])
    return east, north, up


def vincdir(lat1, long1, azimuth1to2, ell_dist, ellipsoid=grs80):
    """
    Vincentys Direct Formula
    :param lat1: Latitude of Point 1 (Decimal Degrees)
    :param long1: Longitude of Point 1 (Decimal Degrees)
    :param azimuth1to2: Azimuth from Point 1 to 2 (Decimal Degrees)
    :param ell_dist: Ellipsoidal Distance between Points 1 and 2 (m)
    :param ellipsoid: Ellipsoid Object
    :return: Latitude of Point 2 (Decimal Degrees),
             Longitude of Point 2 (Decimal Degrees),
             Azimuth from Point 2 to 1 (Decimal Degrees)
    """
    azimuth1to2 = radians(azimuth1to2)

    # Equation numbering is from GDA2020 Tech Manual v1.0
    # Eq. 88
    tan_u1 = (1 - ellipsoid.f) * tan(radians(lat1))
    u1 = atan(tan_u1)
    sin_u1 = sin(u1)
    cos_u1 = cos(u1)
    # Eq. 89
    tan_sigma1 = tan_u1 / cos(azimuth1to2)
    sigma1 = atan(tan_sigma1)
    # Eq. 90
    sin_alpha = cos_u1 * sin(azimuth1to2)
    alpha = asin(sin_alpha)
    # Eq. 91
    u2 = ((cos(alpha)) ** 2
          * (ellipsoid.semimaj ** 2 - ellipsoid.semimin ** 2)
          / ellipsoid.semimin ** 2)
    # Eq. 92
    A = (1 + (u2 / 16384)
         * (4096 + u2
            * (-768 + u2
               * (320 - 175 * u2))))
    # Eq. 93
    B = ((u2 / 1024)
         * (256 + u2
            * (-128 + u2
               * (74 - 47 * u2))))
    # Eq. 94
    sigma = ell_dist / (ellipsoid.semimin * A)
    # Sigma Iteration
    while True:
        # Eq. 95
        sigm2 = 2 * sigma1 + sigma
        # Eq. 96
        sigma_change = (B * sin(sigma) *
                        (cos(sigm2)
                         + (B / 4)
                         * (cos(sigma)
                            * (-1 + 2 * (cos(sigm2)) ** 2)
                            - (B / 6)
                            * cos(sigm2)
                            * (-3 + 4 * sin(sigma) ** 2)
                            * (-3 + 4 * cos(sigm2) ** 2)
                            )))
        # Eq. 97
        sigma = (ell_dist / (ellipsoid.semimin * A)) + sigma_change
        if abs(sigma_change) < 1e-5:
            break
    sin_sigma = sin(sigma)
    cos_sigma = cos(sigma)

    # Calculate Latitude of Pt. 2
    # Eq. 98
    lat2 = atan2((sin_u1 * cos_sigma + cos_u1 * sin(sigma) * cos(azimuth1to2)),
                 ((1 - ellipsoid.f)
                  * sqrt(sin(alpha) ** 2
                         + (sin_u1
                            * sin(sigma)
                            - cos_u1
                            * cos_sigma
                            * cos(azimuth1to2)) ** 2)
                  ))
    lat2 = degrees(lat2)

    # Calculate Longitude of Pt. 2
    # Eq. 99
    long = atan2((sin_sigma * sin(azimuth1to2)),
                 (cos_u1 * cos_sigma - sin_u1 * sin_sigma * cos(azimuth1to2)))
    # Eq. 100
    C = (ellipsoid.f / 16) * cos(alpha) ** 2 * (4 + ellipsoid.f * (4 - 3 * cos(alpha) ** 2))
    # Eq. 101
    omega = (long
             - (1 - C)
             * ellipsoid.f
             * sin_alpha
             * (sigma + C * sin_sigma
                * (cos(sigm2) + C * cos(sigma) * (-1 + 2 * cos(sigm2) ** 2))
                ))
    # Eq. 102
    long2 = float(long1) + degrees(omega)

    # Calculate Reverse Azimuth
    azimuth2to1 = atan(sin_alpha / (-sin_u1 * sin_sigma + cos_u1 * cos_sigma * cos(azimuth1to2))) + pi
    azimuth2to1 = degrees(azimuth2to1)
    return round(lat2, 11), round(long2, 11), round(azimuth2to1, 9)


def vincinv(lat1, long1, lat2, long2, ellipsoid=grs80):
    """
    Vincenty's Inverse Formula
    :param lat1: Latitude of Point 1 (Decimal Degrees)
    :param long1: Longitude of Point 1 (Decimal Degrees)
    :param lat2: Latitude of Point 2 (Decimal Degrees)
    :param long2: Longitude of Point 2 (Decimal Degrees)
    :param ellipsoid: Ellipsoid Object
    :return: Ellipsoidal Distance between Points 1 and 2 (m),
             Azimuth from Point 1 to 2 (Decimal Degrees),
             Azimuth from Point 2 to 1 (Decimal Degrees)
    """
    if lat1 == lat2 and long1 == long2:
        return 0, 0, 0
    # Equation numbering is from GDA2020 Tech Manual v1.0
    # Eq. 71
    tan_u1 = (1 - ellipsoid.f) * tan(radians(lat1))
    u1 = atan(tan_u1)
    sin_u1 = sin(u1)
    cos_u1 = cos(u1)
    # Eq. 72
    tan_u2 = (1 - ellipsoid.f) * tan(radians(lat2))
    u2 = atan(tan_u2)
    sin_u2 = sin(u2)
    cos_u2 = cos(u2)
    # Initial Approximation
    # Eq. 73
    long_diff = radians(long2 - long1)
    omega = long_diff
    sigma = 1
    sigma_diff = 0.5
    itercount = 0
    # Sigma Iteration
    while sigma_diff > 1e-10 and itercount < 100:
        # Eq. 74
        sin_sq_sig = ((cos_u2 * sin(long_diff)) ** 2
                      + (cos_u1 * sin_u2 - sin_u1 * cos_u2 * cos(long_diff)) ** 2)
        sin_sigma = sqrt(sin_sq_sig)
        # Eq. 75
        cos_sigma = sin_u1 * sin_u2 + cos_u1 * cos_u2 * cos(long_diff)
        # Eq. 76
        tan_sigma = sin_sigma / cos_sigma
        # Iterator
        sigma_diff = abs(sigma - atan(tan_sigma))
        sigma = atan(tan_sigma)
        # Eq. 77
        sin_alpha = (cos_u1 * cos_u2 * sin(long_diff)) / sin_sigma
        alpha = asin(sin_alpha)
        # Eq. 78
        cos_2sigm = cos_sigma - ((2 * sin_u1 * sin_u2) / (cos(alpha) ** 2))
        # Eq. 79
        C = (ellipsoid.f / 16) * (cos(alpha) ** 2) * (4 + ellipsoid.f * (4 - 3 * (cos(alpha) ** 2)))
        # Eq. 80
        long_diff = (omega + (1 - C)
                     * ellipsoid.f * sin_alpha
                     * (sigma + C * sin_sigma
                        * (cos_2sigm + C * sigma
                           * (-1 + 2 * cos_2sigm ** 2))))
        itercount += 1

    # Eq. 81
    u_sq = (cos(alpha) ** 2 * (ellipsoid.semimaj ** 2 - ellipsoid.semimin ** 2)) / ellipsoid.semimin ** 2
    # Eq. 82
    A = 1 + (u_sq / 16384) * (4096 + u_sq * (-768 + u_sq * (320 - 175 * u_sq)))
    # Eq. 83
    B = (u_sq / 1024) * (256 + u_sq * (-128 + u_sq * (74 - 47 * u_sq)))
    # Eq. 84
    delta_sigma = (B * sin_sigma
                   * (cos_2sigm + (B / 4)
                      * (cos_sigma * (-1 + 2 * cos_2sigm ** 2)
                         - (B / 6) * cos_2sigm * (-3 + 4 * sin_sigma ** 2)
                         * (-3 + 4 * cos_2sigm ** 2))))
    # Calculate Distance
    if sigma < 0:
        sigma = sigma + pi
    # Eq. 85
    ell_dist = ellipsoid.semimin * A * (sigma - delta_sigma)
    # Calculate Alpha1
    azimuth1to2 = degrees(atan2((cos_u2 * sin(long_diff)),
                                (cos_u1 * sin_u2 - sin_u1
                                 * cos_u2 * cos(long_diff))))
    if azimuth1to2 < 0:
        azimuth1to2 = azimuth1to2 + 360
    # Calculate Alpha2
    azimuth2to1 = degrees(atan2((cos_u1 * sin(long_diff)),
                                (-sin_u1 * cos_u2 + cos_u1
                                 * sin_u2 * cos(long_diff))))
    azimuth2to1 = azimuth2to1 + 180
    # Meridian Critical Case Tests
    if long1 == long2 and lat1 > lat2:
        return ell_dist, 180, 0
    if long1 == long2 and lat1 < lat2:
        return ell_dist, 0, 180
    return round(ell_dist, 3), round(azimuth1to2, 9), round(azimuth2to1, 9)


def vincdirio():
    """
    No Input:
    Prompts the user for the name of a file in csv format. Data in the file
    must be in the form Latitude, Longitude of Point 1 in Degrees Minutes Seconds,
    Geodetic Azimuth from Point 1 to 2 in Degrees Minutes Seconds and Distance
    in metres with no header line.

    No Output:
    Uses the function vincdir to calculate for each row in the csv file the
    geographic coordinate (lat, long) of Point 2 and the Azimuth from Point 2
    to Point 1, all in Degrees Minutes Seconds. This data is written to a new
    file with the name <inputfile>_out.csv
    """
    # Enter Filename
    fn = input('Enter co-ordinate file:\n')
    # Open Filename
    csvfile = open(fn)
    csvreader = csv.reader(csvfile)
    # Create Output File
    fn_part = (os.path.splitext(fn))
    fn_out = fn_part[0] + '_out' + fn_part[1]
    outfile = open(fn_out, 'w')
    # Write Output
    outfilewriter = csv.writer(outfile)
    # outfilewriter.writerow(['Latitude2', 'Longitude2', 'azimuth2to1'])
    for row in csvreader:
        lat1 = dms2dd(float(row[0]))
        long1 = dms2dd(float(row[1]))
        azimuth1to2 = dms2dd(float(row[2]))
        ell_dist = float(row[3])
        lat2, long2, azimuth2to1 = vincdir(lat1, long1, azimuth1to2, ell_dist)
        lat2 = dd2dms(lat2)
        long2 = dd2dms(long2)
        azimuth2to1 = dd2dms(azimuth2to1)
        output = [lat2, long2, azimuth2to1]
        outfilewriter.writerow(output)
    # Close Files
    outfile.close()
    csvfile.close()


def vincinvio():
    # Enter Filename
    print('Enter co-ordinate file:')
    fn = input()
    # Open Filename
    csvfile = open(fn)
    csvreader = csv.reader(csvfile)
    # Create Output File
    fn_part = (os.path.splitext(fn))
    fn_out = fn_part[0] + '_out' + fn_part[1]
    outfile = open(fn_out, 'w')
    # Write Output
    outfilewriter = csv.writer(outfile)
    outfilewriter.writerow(['Ell_Dist', 'Azimuth1to2', 'Azimuth2to1'])
    for row in csvreader:
        lat1 = dms2dd(float(row[0]))
        long1 = dms2dd(float(row[1]))
        lat2 = dms2dd(float(row[2]))
        long2 = dms2dd(float(row[3]))
        ell_dist, azimuth1to2, azimuth2to1 = vincinv(lat1, long1, lat2, long2)
        azimuth1to2 = dd2dms(azimuth1to2)
        azimuth2to1 = dd2dms(azimuth2to1)
        output = (ell_dist, azimuth1to2, azimuth2to1)
        outfilewriter.writerow(output)
    # Close Files
    outfile.close()
    csvfile.close()




#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Conversions Module
"""

from math import modf


def dec2hp(lon, lat):
    """Convert decimal degrees to HP notation

    Longitudes go from -180 to 180
    Latitudes go from -90 to 90
    """
    def fmt_hp(coord):
        """Function to piece together a coordinate in HP notation
        """
        coord = float(coord)
        if coord < 0:
            flag = -1
            coord = abs(coord)
        else:
            flag = 1
        deg = int(coord)
        min = int(60 * (coord - deg))
        sec = 60 * (60 * (coord - deg) - min)
        # Move decimal over two places and round to an integer
        sec = round(sec * 100)
        deg *= flag  # deal with negatives
        hp_coord = '{}.{:02d}{:04d}'.format(deg, min, sec)

        return hp_coord

# Convert the coordinates
    hp_lon = fmt_hp(lon)
    hp_lat = fmt_hp(lat)

    return hp_lon, hp_lat


def hp2dec(lon, lat):
    """Convert HP notation to decimal degrees

    Longitudes go from -180 to 180
    Latitudes go from -90 to 90
    """
    def fmt_dec(coord):
        """Function to piece together a decimal coordinate
        """
        coord = float(coord)
        if coord < 0:
            flag = -1
            coord = abs(coord)
        else:
            flag = 1
        min_sec, deg = modf(coord)
        sec, min = modf(min_sec * 100)
        sec *= 100
        dec_coord = deg + min / 60 + sec / 3600
        dec_coord = float('{:.6f}'.format(dec_coord))
        dec_coord *= flag  # deal with negatives

        return dec_coord

# Convert the coordinates
    dec_lon = fmt_dec(lon)
    dec_lat = fmt_dec(lat)

    return dec_lon, dec_lat


def dec2sex(lon, lat):
    """Convert decimal degrees to sexagesimal format

    Longitudes go from -180 to 180
    Latitudes go from -90 to 90
    """
    def fmt_sex(coord):
        """Function to create a sexagesimal-formatted coordinate as a string
        """
        coord = float(coord)
        if coord < 0:
            flag = -1
            coord = abs(coord)
        else:
            flag = 1
        min_sec, deg = modf(coord)
        deg = int(deg)
        deg *= flag  # deal with negatives
        sec, min = modf(min_sec * 60)
        min = round(min)
        sec *= 60
        sex_coord = '{} {:02d} {:05.2f}'.format(deg, min, sec)

        return sex_coord

# Convert the coordinates
    sex_lon = fmt_sex(lon)
    sex_lat = fmt_sex(lat)

    return sex_lon, sex_lat


def sex2dec(lon, lat):
    """Convert a sexagesimal coordinate to decimal degrees

    Longitudes go from -180 to 180
    Latitudes go from -90 to 90
    """
    def fmt_dec(coord):
        """Function to piece together a decimal coordinate
        """
        coord = str(coord)
        if coord[:1] == '-':
            flag = -1
            coord = coord[1:]
        else:
            flag = 1
        deg, min, sec = coord.split()
        dec_coord = int(deg) + float(min) / 60 + float(sec) / 3600
        dec_coord = float('{:.6f}'.format(dec_coord))
        dec_coord *= flag  # deal with negatives

        return dec_coord

# Convert the coordinates
    dec_lon = fmt_dec(lon)
    dec_lat = fmt_dec(lat)

    return dec_lon, dec_lat


def sex2hp(lon, lat):
    """Convert a sexagesimal coordinate to HP notation

    Longitudes go from -180 to 180
    Latitudes go from -90 to 90
    """
    def fmt_hp(coord):
        """Function to piece together a coordinate in HP notation
        """
        coord = str(coord)
        if coord[:1] == '-':
            flag = -1
            coord = coord[1:]
        else:
            flag = 1
        deg, min, sec = coord.split()
        sec = sec.replace('.', '')
        hp_coord = '{}.{:02d}{:04d}'.format(deg, int(min), int(sec))
        if flag == -1:
            hp_coord = '-' + hp_coord  # deal with negatives

        return hp_coord

# Convert the coordinates
    hp_lon = fmt_hp(lon)
    hp_lat = fmt_hp(lat)

    return hp_lon, hp_lat


def hp2sex(lon, lat):
    """Convert HP notation to a sexagesimal coordinate

    Longitudes go from -180 to 180
    Latitudes go from -90 to 90
    """
    def fmt_sex(coord):
        """Function to piece together a coordinate in HP notation
        """
        coord = str(coord)
        if coord[:1] == '-':
            flag = -1
            coord = coord[1:]
        else:
            flag = 1
        deg, min, sec = coord.split()
        sec = sec.replace('.', '')
        hp_coord = '{}.{:02d}{:04d}'.format(deg, int(min), int(sec))
        if flag == -1:
            hp_coord = '-' + hp_coord  # deal with negatives

        return hp_coord

# Convert the coordinates
    hp_lon = fmt_sex(lon)
    hp_lat = fmt_sex(lat)

    return hp_lon, hp_lat


def dd2dms_v(dd):
    minutes, seconds = divmod(abs(dd) * 3600, 60)
    degrees, minutes = divmod(minutes, 60)
    dms = degrees + (minutes / 100) + (seconds / 10000)
    dms[dd <= 0] = -dms[dd <= 0]
    return dms

def dd2dms(dd):
    minutes, seconds = divmod(abs(dd) * 3600, 60)
    degrees, minutes = divmod(minutes, 60)
    dms = degrees + (minutes / 100) + (seconds / 10000)
    return dms if dd >= 0 else -dms


def dms2dd_v(dms):
    degmin, seconds = divmod(abs(dms) * 1000, 10)
    degrees, minutes = divmod(degmin, 100)
    dd = degrees + (minutes / 60) + (seconds / 360)
    dd[dms <= 0] = -dd[dms <= 0]
    return dd


def dms2dd(dms):
    degmin, seconds = divmod(abs(dms) * 1000, 10)
    degrees, minutes = divmod(degmin, 100)
    dd = degrees + (minutes / 60) + (seconds / 360)
    return dd if dms >= 0 else -dd


class Angle(object):
    def __init__(self, decdeg):
        try:
            self.decdeg = float(decdeg)
        except ValueError:
            print('ValueError: Angle must be numeric')

    def __repr__(self):
        return '{Decimal Angle: ' + str(self.decdeg) + '}'

    def __add__(self, other):
        return self.decdeg + float(other)

    def __sub__(self, other):
        return self.decdeg - float(other)

    def __abs__(self):
        return -self.decdeg if self.decdeg < 0 else self.decdeg

    def __float__(self):
        return self.decdeg

    def hp(self):
        minutes, seconds = divmod(abs(self) * 3600, 60)
        degrees, minutes = divmod(minutes, 60)
        hp = degrees + (minutes / 100) + (seconds / 10000)
        return hp if float(self) >= 0 else -hp



#!/usr/bin/env python3

"""
Geoscience Australia - GeodePy Package
File input/output module

In Development
"""


def rd_snx_est(file):
    """This function reads in the SOLUTION/ESTIMATE block of the given SINEX
    file. It returns solEstimate, a list of tuples:

    solEstimate = [(code, soln, refEpoch, staX, staY, staZ, staX_sd, staY_sd,
                    staZ_sd, velX, velY, velZ, velX_sd, velY_sd, velZ_sd)...]

    where:
        * code is the stations's 4-character ID
        * soln is the segment of the stations's time series
        * refEpoch is the epoch of the solution in the form YY:DOY:SSSSS (YY
            is the two digit year, DOY is day of year, and SSSSS is the time of
            day in seconds
        * sta[XYZ] is the station coordinates in the Cartesian reference frame
        * sta[XYZ]_sd
        * vel[XYZ] is the station velocity in the Cartesian reference frame
        * vel[XYZ]_sd

    :param file: the input SINEX file
    :return: solEstimate
    """

    # Open and read in the given SINEX file
    with open(file) as f:
        lines = f.readlines()

    # Read the solution estimates into a list
    solEstimate = []
    go = False
    for line in lines:
        if line[:18] == '-SOLUTION/ESTIMATE':
            go = False
        if go:
            if component == 'BOTH':
                solEstimate.append(tuple(line.split()))
            else:
                if line[7:10] == component:
                    solEstimate.append(tuple(line.split()))
        if line[:11] == '*INDEX TYPE':
            go = True
    return solEstimate


#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Survey Module
"""

import os
from math import sqrt, degrees, radians, sin, cos, asin
from geodepy.conversions import dd2dms, dms2dd


# Defines a bunch of classes required to convert GSI (or any other format) to DynaNet v3 Format


class AngleObs(object):
    def __init__(self, degree=0, minute=0, second=0):
        self.degree = abs(int(degree))
        self.minute = abs(int(minute))
        self.second = abs(round(float(second), 3))

    def __repr__(self):
        return repr(self.degree) + 'd ' + repr(self.minute) + '\' ' + repr(self.second) + '\"'

    def decimal(self):
        dd = abs(self.degree + self.minute / 60 + self.second / 3600)
        return dd if self.degree >= 0 else -dd

    def hp(self):
        hp = abs(self.degree + self.minute / 100 + self.second / 10000)
        return hp if self.degree >= 0 else -hp

    def __add__(self, other):
        degreeadd = self.degree + other.degree
        minuteadd = abs(self.minute) + abs(other.minute)
        secondadd = abs(self.second) + abs(other.second)
        while secondadd >= 60:
            minuteadd += 1
            secondadd -= 60
        while minuteadd >= 60:
            degreeadd += 1
            minuteadd -= 60
        return AngleObs(degreeadd, minuteadd, secondadd)

    def __sub__(self, other):
        degreesub = self.degree - other.degree
        minutesub = abs(self.minute) - abs(other.minute)
        secondsub = abs(self.second) - abs(other.second)
        while secondsub < 0:
            secondsub += 60
            minutesub -= 1
        while minutesub < 0:
            minutesub += 60
            degreesub -= 1
        return AngleObs(degreesub, minutesub, secondsub)

    def __truediv__(self, other):
        degreediv, degreerem = divmod(self.degree, other)
        minutediv, minuterem = divmod(self.minute, other)
        seconddiv, secondrem = divmod(self.second, other)
        minutediv = minutediv + ((degreerem / other) * 60)
        seconddiv = seconddiv + ((minuterem / other) * 60) + (secondrem / other)
        return AngleObs(degreediv, minutediv, seconddiv)


class Coordinate(object):
    def __init__(self, pt_id, system, hz_datum, vert_datum, epoch, x=0.0, y=0.0, z=0.0):
        self.pt_id = pt_id
        self.system = system
        self.hz_datum = hz_datum
        self.vert_datum = vert_datum
        self.epoch = epoch
        self.x = x
        self.y = y
        self.z = z

    def __repr__(self):
        return ('{' + str(self.hz_datum) + ', '
                + str(self.x) + ', '
                + str(self.y) + ', '
                + str(self.z) + '}')


class InstSetup(object):
    def __init__(self, pt_id, coordinate, observation=None):
        self.pt_id = pt_id
        self.coordinate = coordinate
        if observation is None:
            self.observation = []
        else:
            self.observation = []
            self.observation.append(observation)

    def __repr__(self):
        return ('{InstSetup: ' + repr(self.pt_id)
                + ' ' + repr(self.coordinate)
                + '}\n Observations:\n'
                + repr(self.observation))

    def addobs(self, observation):
        return self.observation.append(observation)

    def __iter__(self):
        return self


class Observation(object):
    def __init__(self, from_id, to_id, inst_height=0.0, target_height=0.0,
                 hz_obs=AngleObs(0, 0, 0), va_obs=AngleObs(0, 0, 0), sd_obs=0.0, hz_dist=0.0, vert_dist=0.0):
        self.from_id = from_id
        self.to_id = to_id
        self.inst_height = inst_height
        self.target_height = target_height
        self.hz_obs = hz_obs
        self.va_obs = va_obs
        self.sd_obs = sd_obs
        self.hz_dist = hz_dist
        self.vert_dist = vert_dist

    def __repr__(self):
        return ('{from: ' + repr(self.from_id)
                + 'to: ' + repr(self.to_id)
                + '; target_height ' + repr(self.target_height)
                + '; hz_obs ' + repr(self.hz_obs)
                + '; va_obs ' + repr(self.va_obs)
                + '; sd_obs ' + repr(self.sd_obs)
                + '; target_height ' + repr(self.target_height)
                + '}')


# Functions to read in data to classes from Leica GSI format file (GA_Survey2.frt)


def readgsi(filepath):
    """
    Takes in a gsi file (GA_Survey2.frt) and returns a list
    of stations with their associated observations.
    :param filepath:
    :return:
    """
    # Check file extension, throw except if not .gsi
    ext = os.path.splitext(filepath)[-1].lower()
    try:
        if ext != '.gsi':
            raise ValueError
    except ValueError:
        print('ValueError: file must have .gsi extension')
        return
    # Read data from gsi file
    with open(filepath, 'r') as file:
        gsidata = file.readlines()
        stn_index = [0]
        for i in gsidata:
            # Create list of line numbers of station records (Only stations have '84..' string)
            if '84..' in i:
                lnid = i[3:7]
                lnid = int(lnid.lstrip('0'))
                stn_index.append(lnid)
        gsi_listbystation = []
        # Create lists of gsi data with station records as first element
        for i in range(0, (len(stn_index) - 1)):
            gsi_listbystation.append(gsidata[(stn_index[i]) - 1:(stn_index[i + 1])])
        del gsi_listbystation[0]
    return gsi_listbystation


def gsi2class(gsi_list):
    """
    Takes a list where first entry is station record and
    all remaining records are observations and creates
    a InstSetup Object with Observation Objects included.
    :param gsi_list:
    :return:
    """
    def readgsiword16(linestring, word_id):
        wordstart = str.find(linestring, word_id)
        word_val = linestring[(wordstart + 7):(wordstart + 23)]
        word_val = word_val.lstrip('0')
        if word_val == '':
            return 0
        else:
            return int(word_val)

    def parse_ptid(gsi_line):
        ptid = gsi_line[8:24]
        ptid = ptid.lstrip('0')
        return ptid

    def parse_easting(gsi_line):
        return (readgsiword16(gsi_line, '84..')) / 10000

    def parse_northing(gsi_line):
        return (readgsiword16(gsi_line, '85..')) / 10000

    def parse_elev(gsi_line):
        return (readgsiword16(gsi_line, '86..')) / 10000

    def parse_hz(gsi_line):
        dms = readgsiword16(gsi_line, '21.324') / 100000
        degmin, second = divmod(abs(dms) * 1000, 10)
        degree, minute = divmod(degmin, 100)
        return AngleObs(degree, minute, second * 10)

    def parse_slope(gsi_line):
        return (readgsiword16(gsi_line, '31..')) / 10000

    def parse_dist(gsi_line):
        return (readgsiword16(gsi_line, '32..')) / 10000

    def parse_tgtht(gsi_line):
        return (readgsiword16(gsi_line, '87..')) / 10000

    def parse_instht(gsi_line):
        return (readgsiword16(gsi_line, '88..')) / 10000

    def parse_vert(gsi_line):
        dms = readgsiword16(gsi_line, '22.324') / 100000
        degmin, seconds = divmod(abs(dms) * 1000, 10)
        degrees, minutes = divmod(degmin, 100)
        return AngleObs(degrees, minutes, seconds * 10)

    project = []
    for record in gsi_list:
        from_stn = parse_ptid(record[0])
        obs_list = []
        for line in record:
            if '31..' in line:
                to_stn = parse_ptid(line)
                hz = parse_hz(line)
                vert = parse_vert(line)
                slope = parse_slope(line)
                tgtht = parse_tgtht(line)
                instht = parse_instht(line)
                obs = Observation(from_stn, to_stn, instht, tgtht, hz, vert, slope)
                obs_list.append(obs)
        if '84..' in record[0]:
            pt_id = parse_ptid(record[0])
            easting = parse_easting(record[0])
            northing = parse_northing(record[0])
            elev = parse_elev(record[0])
            coord = Coordinate(pt_id, 'utm', 'gda94', 'gda94', '2018.1', easting, northing, elev)
            setup = InstSetup(pt_id, coord)
            for i in range(0, len(obs_list)):
                setup.addobs(obs_list[i])
        project.append(setup)
    return project


def readgsiword16(linestring, word_id):
    wordstart = str.find(linestring, word_id)
    word_val = linestring[(wordstart + 7):(wordstart + 23)]
    word_val = 0.0 if word_val.lstrip('0') == '' else int(word_val.lstrip('0'))
    return word_val


# Functions to write out station and obs data to DNA format

def anglerounds(setup):
    """

    :param setup:
    :return:
    """
    def meanhz(angle1, angle2):
        if angle1.degree < angle2.degree:
            return (angle1 + (angle2 - AngleObs(180))) / 2
        else:
            return (angle1 + (angle2 + AngleObs(180))) / 2

    def meanva(angle1, angle2):
        if angle1.degree < 180:
            return (angle1 + (AngleObs(360) - angle2)) / 2
        else:
            return (angle2 + (AngleObs(360) - angle1)) / 2
    # Build a round of FL/FR obs
    start_round = setup.observation[0].to_id
    ind = []
    roundofobs = []
    for c, ob in enumerate(setup.observation):
        if ob.to_id == start_round:
            ind.append(c)
    roundofobs = setup.observation[ind[0]:ind[1] + 1]
    # Test for Even Palindromic to_stn names
    meanround = []
    fl = 0
    fr = -1
    while len(roundofobs)/2 > fl + 1:
        meanhz = meanhz(roundofobs[fl].hz_obs, roundofobs[fr].hz_obs)
        meanva = meanva(roundofobs[fr].va_obs, roundofobs[fr].va_obs)
        meanob = Observation(roundofobs[fl].from_id,
                             roundofobs[fl].to_id,
                             roundofobs[fl].inst_height,
                             roundofobs[fl].target_height,
                             meanhz,
                             meanva,
                             roundofobs[fl].sd_obs,
                             roundofobs[fl].hz_dist,
                             roundofobs[fl].vert_dist)
        meanround.append(meanob)
        fl += 1
        fr -= 1
    # Create new list of meaned rounds of obs
    # For first and last in round
        # mean_va = (FL_va + FR_hz) / 2 (tbd)
        # mean_sd = (FL_sd + FL_sd) / 2
        # Observation(All same + mean_hz, mean_va, mean_sd)
        # Add above to list of meaned rounds of obs
        # Remove first and last in round
    # Repeat until all obs in round meaned
    # Repeat above until all rounds meaned
    # Return new list of meaned obs per round (no non-round obs included)
    pass


def anglepairs(allobsfromInstSetup):
    # For first to_stn
        # Find next occurrence of to_stn
        # Check (FL_hz ~= (FR_hz +- 180)
    # Create new list of meaned rounds of obs
    # For FL/FR pair of obs
    def meanhz(angle1, angle2):
        if angle1.degree < angle2.degree:
            return (angle1 + (angle2 - AngleObs(180))) / 2
        else:
            return (angle1 + (angle2 + AngleObs(180))) / 2

    def meanva(angle1, angle2):
        if angle1.degree < 180:
            return (angle1 + (AngleObs(360) - angle2)) / 2
        else:
            return (angle2 + (AngleObs(360) - angle1)) / 2
        # mean_sd = (FL_sd + FL_sd) / 2
        # Observation(All same + mean_hz, mean_va, mean_sd)
        # Remove pair from list
    # Repeat until all pairs of obs in round meaned
    # Return new list of meaned pairs of obs (no single-face obs included)
    pass


def dnaout_dirset(allobsfromInstSetup):
    # Find index of next obs with same to_stn (defines direction set)
    # If no next, include all obs
    # Set default std_dev
    # For obs in unique list
        # First line is first obs
        # For remaining obs
            # Include to_stn and direction
    pass


def dnaout_sd(observation):
    return ('S '
            + observation.from_id.ljust(20)
            + observation.to_id.ljust(20)
            + ''.ljust(20)
            + str(observation.sd_obs).ljust(14)  # 76
            + ''.ljust(14)
            + '0.0010'.ljust(9)         # add standard deviation
            + str(1.7960).ljust(7)      # add intrument height
            + str(observation.target_height).ljust(7))


def dnaout_va(observation):
    return ('V '
            + observation.from_id.ljust(20)
            + observation.to_id.ljust(20)
            + ''.ljust(34)
            + str(observation.va_obs.degree).rjust(3)
            + ' '
            + str('%02d' % observation.va_obs.minute)
            + ' '
            + str(observation.va_obs.second).ljust(8)
            + '1.0000'.ljust(9)         # add standard deviation
            + str(1.7960).ljust(7)      # add intrument height
            + str(observation.target_height).ljust(7))


def va_conv(verta_hp, slope_dist, height_inst=0, height_tgt=0):
    """
    Function to convert vertical angles (zenith distances) and slope distances into horizontal
    distances and changes in height. Instrument and Target heights can be entered to allow
    computation of zenith and slope distances between ground points.

    :param verta_hp:        Vertical Angle from Instrument to Target, expressed in HP Format (DDD.MMSSSSSS)
    :param slope_dist:      Slope Distance from Instrument to Target in metres
    :param height_inst:     Height of Instrument. Optional - Default Value of 0m
    :param height_tgt:      Height of Target. Optional - Default Value of 0m

    :return: verta_pt_hp:   Vertical Angle between Ground Points, expressed in HP Format (DDD.MMSSSSSS)
    :return: slope_dist_pt: Slope Distance between Ground Points in metres
    :return: hz_dist:       Horizontal Distance
    :return: delta_ht:      Change in height between Ground Points in metres
    """
    # Convert Zenith Angle to Vertical Angle
    try:
        if verta_hp == 0 or verta_hp == 180:
            raise ValueError
        elif 0 < verta_hp < 180:
            verta = radians(90 - dms2dd(verta_hp))
        elif 180 < verta_hp < 360:
            verta = radians(270 - dms2dd(verta_hp))
        else:
            raise ValueError
    except ValueError:
        print('ValueError: Vertical Angle Invalid')
        return
    # Calculate Horizontal Dist and Delta Height
    hz_dist = slope_dist * cos(verta)
    delta_ht = slope_dist * sin(verta)
    # Account for Target and Instrument Heights
    if height_inst == 0 and height_tgt == 0:
        verta_pt_hp = verta_hp
        slope_dist_pt = slope_dist
    else:
        delta_ht = height_inst + delta_ht - height_tgt
        slope_dist_pt = sqrt(delta_ht ** 2 + hz_dist ** 2)
        verta_pt = asin(delta_ht / slope_dist)
        verta_pt_hp = dd2dms(degrees(verta_pt) + 90)
    return verta_pt_hp, slope_dist_pt, hz_dist, delta_ht


"""
to_stn = ['GA03', 'GA02', 'SYM2', 'ATWR', 'MAHON', 'MAHON', 'ATWR', 'SYM2', 'GA02', 'GA03']
flfr = [359.59597, 14.48289, 25.35515, 215.57043, 300.59046, 120.59060, 35.57045, 205.35530, 194.48282, 179.59581]
flfr_vert = [90.30228, 90.09547, 89.50396, 88.03600, 87.58014, 272.01587, 271.55593, 270.09227, 269.50091, 269.29414]
"""


def hz_round(brg_list):
    """
    Input: an even palindromic list of horizontal angle observations in two faces
    (e.g. [fl1, fl2, fl3, fr3, fr2, fr1])
    Output: an averaged face-left sense list of horizontal angles.
    """
    brg_avg = []
    obs = int((len(brg_list))/2)
    for i in range(0, obs):
        hz_avg = (dms2dd(brg_list[i]) + (dms2dd(brg_list[-(i+1)])-180))/2
        brg_avg.append(round(dd2dms(hz_avg), 7))
    return brg_avg


def va_round(va_list):
    """
    Input: an even palindromic list of vertical angle observations in two faces
    (e.g. [fl1, fl2, fl3, fr3, fr2, fr1])
    Output: an averaged face-left sense list of vertical angles.
    """
    va_avg = []
    obs = int((len(va_list))/2)
    for i in range(0, obs):
        fl_ang = dms2dd(va_list[i]) - 90
        fr_ang = 270 - dms2dd(va_list[-(i+1)])
        ang_avg = (fl_ang + fr_ang)/2 + 90
        va_avg.append(round(dd2dms(ang_avg), 7))
    return va_avg


"""
# Test for round of obs
if to_stn == to_stn[::-1] and len(to_stn) % 2 == 0:
    brg_avg = hz_round(flfr)
    va_avg = va_round(flfr_vert)
    to_stn_avg = to_stn[0:int(len(to_stn)/2)]
else:
    brg_avg = ['nope']
"""



#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Transform Module

Ref1: http://www.icsm.gov.au/sites/default/files/GDA2020TechnicalManualV1.1.1.pdf
Ref2: http://www.mygeodesy.id.au/documents/Karney-Krueger%20equations.pdf
"""


import os
import csv
from math import sqrt, log, degrees, radians, sin, cos, tan, sinh, cosh, atan, atan2, modf
import numpy as np
from geodepy.constants import grs80, utm, Transformation
from geodepy.convert import dd2dms, dms2dd


# Universal Transverse Mercator Projection Parameters
proj = utm


def rect_radius(ellipsoid):
    """
    Computes the Rectifying Radius of an Ellipsoid with specified Inverse Flattening
    (See Ref 2 Equation 3)
    :param ellipsoid: Ellipsoid Object
    :return: Ellipsoid Rectifying Radius
    """
    nval = (1 / float(ellipsoid.inversef)) / (2 - (1 / float(ellipsoid.inversef)))
    nval2 = nval ** 2
    return (ellipsoid.semimaj / (1 + nval) * ((nval2 *
                                              (nval2 *
                                               (nval2 *
                                                (25 * nval2 + 64)
                                                + 256)
                                               + 4096)
                                              + 16384)
                                              / 16384.))


def alpha_coeff(ellipsoid):
    """
    Computes the set of Alpha coefficients of an Ellipsoid with specified Inverse Flattening
    (See Ref 2 Equation 5)
    :param ellipsoid: Ellipsoid Object
    :return: Alpha coefficients a2, a4 ... a16
    """
    nval = ellipsoid.n
    a2 = ((nval *
           (nval *
            (nval *
             (nval *
              (nval *
               (nval *
                ((37884525 - 75900428 * nval)
                 * nval + 42422016)
                - 89611200)
               + 46287360)
              + 63504000)
             - 135475200)
            + 101606400))
          / 203212800.)

    a4 = ((nval ** 2 *
           (nval *
            (nval *
             (nval *
              (nval *
               (nval *
                (148003883 * nval + 83274912)
                - 178508970)
               + 77690880)
              + 67374720)
             - 104509440)
            + 47174400))
          / 174182400.)

    a6 = ((nval ** 3 *
           (nval *
            (nval *
             (nval *
              (nval *
               (318729724 * nval - 738126169)
               + 294981280)
              + 178924680)
             - 234938880)
            + 81164160))
          / 319334400.)

    a8 = ((nval ** 4 *
           (nval *
            (nval *
             ((14967552000 - 40176129013 * nval) * nval + 6971354016)
             - 8165836800)
            + 2355138720))
          / 7664025600.)

    a10 = ((nval ** 5 *
            (nval *
             (nval *
              (10421654396 * nval + 3997835751)
              - 4266773472)
             + 1072709352))
           / 2490808320.)

    a12 = ((nval ** 6 *
            (nval *
             (175214326799 * nval - 171950693600)
             + 38652967262))
           / 58118860800.)

    a14 = ((nval ** 7 *
            (13700311101 - 67039739596 * nval))
           / 12454041600.)

    a16 = (1424729850961 * nval ** 8) / 743921418240.
    return a2, a4, a6, a8, a10, a12, a14, a16


def beta_coeff(ellipsoid):
    """
    Computes the set of Beta coefficients of an Ellipsoid with specified Inverse Flattening
    (See Ref 2 Equation 23)
    :param ellipsoid: Ellipsoid Object
    :return: Alpha coefficients a2, a4 ... a16
    """
    nval = ellipsoid.n
    b2 = ((nval *
           (nval *
            (nval *
             (nval *
              (nval *
               (nval *
                ((37845269 - 31777436 * nval) - 43097152)
                + 42865200)
               + 752640)
              - 104428800)
             + 180633600)
            - 135475200))
          / 270950400.)

    b4 = ((nval ** 2 *
           (nval *
            (nval *
             (nval *
              (nval *
               ((-24749483 * nval - 14930208) * nval + 100683990)
               - 152616960)
              + 105719040)
             - 23224320)
            - 7257600))
          / 348364800.)

    b6 = ((nval ** 3 *
           (nval *
            (nval *
             (nval *
              (nval *
               (232468668 * nval - 101880889)
               - 39205760)
              + 29795040)
             + 28131840)
            - 22619520))
          / 638668800.)

    b8 = ((nval ** 4 *
           (nval *
            (nval *
             ((-324154477 * nval - 1433121792) * nval + 876745056)
             + 167270400)
            - 208945440))
          / 7664025600.)

    b10 = ((nval ** 5 *
            (nval *
             (nval *
              (312227409 - 457888660 * nval)
              + 67920528)
             - 70779852))
           / 2490808320.)

    b12 = ((nval ** 6 *
            (nval *
             (19841813847 * nval + 3665348512)
             - 3758062126))
           / 116237721600.)

    b14 = ((nval ** 7 *
            (1989295244 * nval - 1979471673))
           / 49816166400.)

    b16 = ((-191773887257 * nval ** 8) / 3719607091200.)
    return b2, b4, b6, b8, b10, b12, b14, b16


def psfandgridconv(xi1, eta1, lat, long, cm, conf_lat, ellipsoid=grs80):
    A = rect_radius(ellipsoid)
    a = alpha_coeff(ellipsoid)
    lat = radians(lat)
    long_diff = radians(long - cm)

    # Point Scale Factor
    p = 1
    q = 0
    for r in range(1, 9):
        p += 2*r * a[r-1] * cos(2*r * xi1) * cosh(2*r * eta1)
        q += 2*r * a[r-1] * sin(2*r * xi1) * sinh(2*r * eta1)
    q = -q
    psf = (float(proj.cmscale)
           * (A / ellipsoid.semimaj)
           * sqrt(q**2 + p**2)
           * ((sqrt(1 + (tan(lat)**2)) * sqrt(1 - ellipsoid.ecc1sq * (sin(lat)**2)))
              / sqrt((tan(conf_lat)**2) + (cos(long_diff)**2))))

    # Grid Convergence
    grid_conv = degrees(atan(abs(q / p))
                        + atan(abs(tan(conf_lat) * tan(long_diff))
                               / sqrt(1 + tan(conf_lat)**2)))
    if cm > long and lat < 0:
        grid_conv = -grid_conv
    elif cm < long and lat > 0:
        grid_conv = -grid_conv

    return psf, grid_conv


def geo2grid(lat, long, zone=0, ellipsoid=grs80):
    """
    Takes a geographic co-ordinate (latitude, longitude) and returns its corresponding
    Hemisphere, Zone and Projection Easting and Northing, Point Scale Factor and Grid
    Convergence. Default Projection is Universal Transverse Mercator Projection using
    GRS80 Ellipsoid parameters.
    :param lat: Latitude in Decimal Degrees
    :param long: Longitude in Decimal Degrees
    :param zone: Optional Zone Number - Only required if calculating grid co-ordinate
                outside zone boundaries
    :param ellipsoid: Ellipsoid Object
    :return: hemisphere, zone, east (m), north (m), Point Scale Factor, Grid Convergence (Decimal Degrees)
    """
    # Input Exception Handling - UTM Extents and Values
    try:
        zone = int(zone)
        if zone < 0 or zone > 60:
            raise ValueError
    except ValueError:
        print('ValueError: Invalid Zone - Zones from 1 to 60')
        return
    try:
        if lat < -80 or lat > 84:
            raise ValueError
    except ValueError:
        print('ValueError: Invalid Latitude - Latitudes from -80 to +84')
        return
    try:
        if long < -180 or long > 180:
            raise ValueError
    except ValueError:
        print('ValueError: Invalid Longitude - Longitudes from -180 to +180')
        return

    A = rect_radius(ellipsoid)
    a = alpha_coeff(ellipsoid)
    lat = radians(lat)
    # Calculate Zone
    if zone == 0:
        zone = int((float(long) - (proj.initialcm - (1.5 * proj.zonewidth))) / proj.zonewidth)
    cm = float(zone * proj.zonewidth) + (proj.initialcm - proj.zonewidth)

    # Conformal Latitude
    sigx = (ellipsoid.ecc1 * tan(lat)) / sqrt(1 + (tan(lat) ** 2))
    sig = sinh(ellipsoid.ecc1 * (0.5 * log((1 + sigx) / (1 - sigx))))
    conf_lat = tan(lat) * sqrt(1 + sig ** 2) - sig * sqrt(1 + (tan(lat) ** 2))
    conf_lat = atan(conf_lat)

    # Longitude Difference
    long_diff = radians(long - cm)
    # Gauss-Schreiber Ratios
    xi1 = atan(tan(conf_lat) / cos(long_diff))
    eta1x = sin(long_diff) / (sqrt(tan(conf_lat) ** 2 + cos(long_diff) ** 2))
    eta1 = log(eta1x + sqrt(1 + eta1x ** 2))

    # Transverse Mercator Ratios
    eta = eta1
    xi = xi1
    for r in range(1, 9):
        eta += a[r-1] * cos(2*r * xi1) * sinh(2*r * eta1)
        xi += a[r-1] * sin(2*r * xi1) * cosh(2*r * eta1)

    # Transverse Mercator Co-ordinates
    x = A * eta
    y = A * xi

    # Hemisphere-dependent UTM Projection Co-ordinates
    east = proj.cmscale * x + proj.falseeast
    if y < 0:
        hemisphere = 'South'
        north = proj.cmscale * y + proj.falsenorth
    else:
        hemisphere = 'North'
        falsenorth = 0
        north = proj.cmscale * y + falsenorth

    # Point Scale Factor and Grid Convergence
    psf, grid_conv = psfandgridconv(xi1, eta1, degrees(lat), long, cm, conf_lat)

    return hemisphere, zone, round(float(east), 3), round(float(north), 3), round(psf, 8), grid_conv


def grid2geo(zone, east, north, hemisphere='south', ellipsoid=grs80):
    """
    Takes a Transverse Mercator grid co-ordinate (Zone, Easting, Northing, Hemisphere)
    and returns its corresponding Geographic Latitude and Longitude, Point Scale Factor
    and Grid Convergence. Default Projection is Universal Transverse Mercator Projection
    using GRS80 Ellipsoid parameters.
    :param zone: Zone Number - 1 to 60
    :param east: Easting (m, within 3330km of Central Meridian)
    :param north: Northing (m, 0 to 10,000,000m)
    :param hemisphere: String - 'North' or 'South'(default)
    :param ellipsoid: Ellipsoid Object
    :return: Latitude and Longitude (Decimal Degrees), Point Scale Factor, Grid Convergence (Decimal Degrees)
    """
    # Input Exception Handling - UTM Extents and Values
    try:
        zone = int(zone)
        if zone < 0 or zone > 60:
            raise ValueError
    except ValueError:
        print('ValueError: Invalid Zone - Zones from 1 to 60')
        return
    try:
        if east < -2830000 or east > 3830000:
            raise ValueError
    except ValueError:
        print('ValueError: Invalid Easting - Must be within 3330km of Central Meridian')
        return
    try:
        if north < 0 or north > 10000000:
            raise ValueError
    except ValueError:
        print('ValueError: Invalid Northing - Must be between 0 and 10,000,000m')
        return
    try:
        h = hemisphere.lower()
        if h != 'north' and h != 'south':
            raise ValueError
    except ValueError:
        print('ValueError: Invalid Hemisphere - String, either North or South')
        return

    A = rect_radius(ellipsoid)
    b = beta_coeff(ellipsoid)
    # Transverse Mercator Co-ordinates
    x = (east - float(proj.falseeast)) / float(proj.cmscale)
    if hemisphere == 'north':
        y = north / float(proj.cmscale)
    else:
        y = (north - float(proj.falsenorth)) / float(proj.cmscale)

    # Transverse Mercator Ratios
    xi = y / A
    eta = x / A

    # Gauss-Schreiber Ratios
    eta1 = eta
    xi1 = xi
    for r in range(1, 9):
        eta1 += b[r-1] * cos(2*r * xi) * sinh(2*r * eta)
        xi1 += b[r-1] * sin(2*r * xi) * cosh(2*r * eta)

    # Conformal Latitude
    conf_lat = (sin(xi1)) / (sqrt((sinh(eta1)) ** 2 + (cos(xi1)) ** 2))
    t1 = conf_lat
    conf_lat = atan(conf_lat)

    # Finding t using Newtons Method
    def sigma(tn, ecc1):
        return (sinh(ecc1
                     * 0.5
                     * log((1 + ((ecc1 * tn) / (sqrt(1 + tn ** 2))))
                           / (1 - ((ecc1 * tn) / (sqrt(1 + tn ** 2)))))))

    def ftn(tn, ecc1):
        return t * sqrt(1 + (sigma(tn, ecc1)) ** 2) - sigma(tn, ecc1) * sqrt(1 + tn ** 2) - t1

    def f1tn(tn, ecc1, ecc1sq):
        return ((sqrt(1 + (sigma(tn, ecc1)) ** 2) * sqrt(1 + tn ** 2) - sigma(tn, ecc1) * tn)
                * (((1 - float(ecc1sq)) * sqrt(1 + t ** 2))
                   / (1 + (1 - float(ecc1sq)) * t ** 2)))

    diff = 1
    t = t1
    itercount = 0
    while diff > 1e-15 and itercount < 100:
        itercount += 1
        t_before = t
        t = t - (ftn(t, ellipsoid.ecc1) / f1tn(t, ellipsoid.ecc1, ellipsoid.ecc1sq))
        diff = abs(t - t_before)
    lat = degrees(atan(t))

    # Compute Longitude
    cm = float((zone * proj.zonewidth) + proj.initialcm - proj.zonewidth)
    long_diff = degrees(atan(sinh(eta1) / cos(xi1)))
    long = cm + long_diff

    # Point Scale Factor and Grid Convergence
    psf, grid_conv = psfandgridconv(xi1, eta1, lat, long, cm, conf_lat)

    return round(lat, 11), round(long, 11), round(psf, 8), grid_conv


def xyz2llh(x, y, z, ellipsoid=grs80):
    # Add input for ellipsoid (default: grs80)
    """
    Input: Cartesian XYZ coordinate in metres

    Output: Latitude and Longitude in Decimal
    Degrees and Ellipsoidal Height in Metres
    """
    # Calculate Longitude
    long = atan2(y, x)
    # Calculate Latitude
    p = sqrt(x**2 + y**2)
    latinit = atan((z*(1+ellipsoid.ecc2sq))/p)
    lat = latinit
    itercheck = 1
    while abs(itercheck) > 1e-10:
        nu = ellipsoid.semimaj/(sqrt(1 - ellipsoid.ecc1sq * (sin(lat))**2))
        itercheck = lat - atan((z + nu * ellipsoid.ecc1sq * sin(lat))/p)
        lat = atan((z + nu * ellipsoid.ecc1sq * sin(lat))/p)
    nu = ellipsoid.semimaj/(sqrt(1 - ellipsoid.ecc1sq * (sin(lat))**2))
    ellht = p/(cos(lat)) - nu
    # Convert Latitude and Longitude to Degrees
    lat = degrees(lat)
    long = degrees(long)
    return lat, long, ellht


def llh2xyz(lat, long, ellht, ellipsoid=grs80):
    # Add input for ellipsoid (default: grs80)
    """
    Input: Latitude and Longitude in Decimal Degrees, Ellipsoidal Height in metres
    Output: Cartesian X, Y, Z Coordinates in metres
    """
    # Convert lat & long to radians
    lat = radians(lat)
    long = radians(long)
    # Calculate Ellipsoid Radius of Curvature in the Prime Vertical - nu
    if lat == 0:
        nu = grs80.semimaj
    else:
        nu = ellipsoid.semimaj/(sqrt(1 - ellipsoid.ecc1sq * (sin(lat)**2)))
    # Calculate x, y, z
    x = (nu + ellht) * cos(lat) * cos(long)
    y = (nu + ellht) * cos(lat) * sin(long)
    z = ((ellipsoid.semimin**2 / ellipsoid.semimaj**2) * nu + ellht) * sin(lat)
    return x, y, z


# # conform7 Debug - Test Values ALIC
# x = -4052051.7643
# y = 4212836.2017
# z = -2545106.0245
# # Debug - Test Helmert Params (GDA94 to GDA2020, scale in ppm and rotations in seconds)
# conform_gda94to20 = [0.06155, -0.01087, -0.04019, -0.009994, -0.0394924, -0.0327221, -0.0328979]


def conform7(x, y, z, trans):
    """
    Performs a Helmert 7 Parameter Conformal Transformation using Cartesian point co-ordinates
    and a predefined transformation object.
    :param x: Cartesian X (m)
    :param y: Cartesian Y (m)
    :param z: Cartesian Z (m)
    :param trans: Transformation Object
    :return: Transformed X, Y, Z Cartesian Co-ordinates
    """
    if type(trans) != Transformation:
        raise ValueError('trans must be a Transformation Object')
    # Create XYZ Vector
    xyz_before = np.array([[x],
                           [y],
                           [z]])
    # Convert Units for Transformation Parameters
    scale = trans.sc / 1000000
    rx = radians(dms2dd(trans.rx / 10000))
    ry = radians(dms2dd(trans.ry / 10000))
    rz = radians(dms2dd(trans.rz / 10000))
    # Create Translation Vector
    translation = np.array([[trans.tx],
                            [trans.ty],
                            [trans.tz]])
    # Create Rotation Matrix
    rotation = np.array([[1., rz, -ry],
                         [-rz, 1., rx],
                         [ry, -rx, 1.]])
    # Conformal Transform Eq
    xyz_after = translation + (1 + scale) * np.dot(rotation, xyz_before)
    # Convert Vector to Separate Variables
    xtrans = float(xyz_after[0])
    ytrans = float(xyz_after[1])
    ztrans = float(xyz_after[2])
    return xtrans, ytrans, ztrans


def conform14(x, y, z, to_epoch, trans):
    """
    Performs a Helmert 14 Parameter Conformal Transformation using Cartesian point co-ordinates
    and a predefined transformation object. The transformation parameters are projected from
    the transformation objects reference epoch to a specified epoch.
    :param x: Cartesian X (m)
    :param y: Cartesian Y (m)
    :param z: Cartesian Z (m)
    :param to_epoch: Epoch co-ordinate transformation is performed at in YYYY.DOY notation
    :param trans: Transformation Object
    :return: Cartesian X, Y, Z co-ordinates transformed using Transformation parameters at desired epoch
    """
    if type(trans) != Transformation:
        raise ValueError('trans must be a Transformation Object')
    # Convert YYYY.DOY to Decimal Year
    to_doy, to_year = modf(to_epoch)
    ref_doy, ref_year = modf(trans.ref_epoch)
    to_epoch = to_year + ((to_doy - 0.0005) / 0.36525)
    ref_epoch = ref_year + (ref_doy / 0.36525)
    # Perform Conformal 7 Parameter Transformation
    # debug - output Transformation Object
    timetrans = (trans + (to_epoch - ref_epoch))
    xtrans, ytrans, ztrans = conform7(x, y, z, timetrans)
    return xtrans, ytrans, ztrans  # , timetrans


def grid2geoio():
    """
    No Input:
    Prompts the user for the name of a file in csv format. Data in the file
    must be in the form Point ID, UTM Zone, Easting (m), Northing (m) with
    no header line.

    No Output:
    Uses the function grid2geo to convert each row in the csv file into a
    latitude and longitude in Degrees, Minutes and Seconds. This data is
    written to a new file with the name <inputfile>_out.csv
    """
    # Enter Filename
    print('Enter co-ordinate file (\.csv)\:')
    fn = input()
    # Open Filename
    csvfile = open(fn)
    csvreader = csv.reader(csvfile)
    # Create Output File
    fn_part = (os.path.splitext(fn))
    fn_out = fn_part[0] + '_out' + fn_part[1]
    outfile = open(fn_out, 'w')
    # Write Output
    outfilewriter = csv.writer(outfile)
    # Optional Header Row
    # outfilewriter.writerow(['Pt', 'Latitude', 'Longitude', 'Point Scale Factor', 'Grid Convergence'])
    for row in csvreader:
        pt_num = row[0]
        zone = float(row[1])
        east = float(row[2])
        north = float(row[3])
        # Calculate Conversion
        lat, long, psf, grid_conv = grid2geo(zone, east, north)
        lat = dd2dms(lat)
        long = dd2dms(long)
        grid_conv = dd2dms(grid_conv)
        output = [pt_num, lat, long, psf, grid_conv]
        outfilewriter.writerow(output)
    # Close Files
    outfile.close()
    csvfile.close()


def geo2gridio():
    """
    No Input:
    Prompts the user for the name of a file in csv format. Data in the file
    must be in the form Point ID, Latitude, Longitude in Decimal Degrees with
    no header line.

    No Output:
    Uses the function geo2grid to convert each row in the csv file into a
    coordinate with UTM Zone, Easting (m), Northing (m). This data is written
    to a new file with the name <inputfile>_out.csv
    """
    # Enter Filename
    print('Enter co-ordinate file:')
    fn = input()
    # Open Filename
    csvfile = open(fn)
    csvreader = csv.reader(csvfile)
    # Create Output File
    fn_part = (os.path.splitext(fn))
    fn_out = fn_part[0] + '_out' + fn_part[1]
    outfile = open(fn_out, 'w')
    # Write Output
    outfilewriter = csv.writer(outfile)
    # Optional Header Row
    # outfilewriter.writerow(['Pt', 'Zone', 'Easting', 'Northing', 'Point Scale Factor', 'Grid Convergence'])
    for row in csvreader:
        pt_num = row[0]
        lat = dms2dd(float(row[1]))
        long = dms2dd(float(row[2]))
        # Calculate Conversion
        hemisphere, zone, east, north, psf, grid_conv = geo2grid(lat, long)
        grid_conv = dms2dd(grid_conv)
        output = [pt_num] + [hemisphere, zone, east, north, psf, grid_conv]
        outfilewriter.writerow(output)
    # Close Files
    outfile.close()
    csvfile.close()
